/*
Original Hires grabber by enthusi/onslaught
multicolor support added by iAN CooG/HVSC in version 2.0
charset/screen support added by enthusi/onslaught in version 3.0
Doodle format picture saving added by TSM in version 4.4
16/02/11
fix: C64MEM pointer was fixed to 0x80, it changed in vice 2.2.x/2.3
fix: text mode, if CHARGEN is used (font at $1x00/$9x00) don't use font
12/12/12
fix: the string "VIC-II:" is found in TFR 0.8 ROM, interferes with the
     actual "VIC-II" block, now it checks for "VIC-II\0\0" and does a
     minimal sanity check, also on "CIA2" block. It's still a kludge but
     should be better than before
19/03/14
new: saves hires pics in Doodle format, too
12/12/19
new: saves editor-friendly executables for PETSCII screens
*/

#include <stdio.h>
#include <string.h>

#define FILENAMELENGTH 256
#define LINE for(i=0;i<79;i++)printf("-");printf("\n")

/* CHANGE THESE FOR OTHER VICE VERSIONS MAYBE */
#define VICOFFSET 0x48d
#define CIAOFFSET 0x16
#define ATTOFFSET 0x41
#define MEMOFFSET 0x1a

/* offset inside basic for $d020/21 values */
#define B_PATCH_D020 0x17

/* offset inside mbasic for $d020/21 values */
#define M_PATCH_D020 0x17
#define M_PATCH_D021 0x1c

/* offset inside fbasic for $d020/21 values */
#define F_PATCH_D020 0x17
#define F_PATCH_D021 0x1c
#define F_PATCH_D022 0x21
#define F_PATCH_D023 0x26
#define F_PATCH_D016 0x2b

/* offset inside pbasic for $d020/21 values */
#define P_PATCH_D020 0x19
#define P_PATCH_D021 0x1e

#define VERSION "4.5"

int is_chargen(int font)
{
    return((font&0x7000)==0x1000);
}

int main (int argc, char **argv)
{
    unsigned char basic[] =
    { 0x01, 0x08, 0x13, 0x08, 0x00, 0x00, 0x9e, 0x32, 0x30, 0x36, 0x39,
      0x20, 0x45, 0x4e, 0x54, 0x48, 0x55, 0x53, 0x49, 0x00, 0x00, 0x00,
      0xa9, 0x00, 0x8d, 0x20, 0xd0, 0xa9, 0xff, 0x8d, 0x00, 0xdd, 0xa9,
      0x3b, 0x8d, 0x11, 0xd0, 0xa9, 0x38, 0x8d, 0x18, 0xd0, 0x4c, 0x29, 0x08
    };
    unsigned char mbasic[] =
    { 0x01, 0x08, 0x13, 0x08, 0x00, 0x00, 0x9e, 0x32,
      0x30, 0x36, 0x39, 0x20, 0x49, 0x41, 0x4e, 0x43,
      0x4f, 0x4f, 0x47, 0x00, 0x00, 0x00, 0xa9, 0x00,  /* <- 0x17 */
      0x8d, 0x20, 0xd0, 0xa9, 0x00, 0x8d, 0x21, 0xd0,  /* <- 0x1c */
      0xa9, 0xff, 0x8d, 0x00, 0xdd, 0xa9, 0x3b, 0x8d,
      0x11, 0xd0, 0xa9, 0x38, 0x8d, 0x18, 0xd0, 0xA9,
      0x18, 0x8D, 0x16, 0xD0, 0xA2, 0x00, 0xBD, 0x00,
      0x10, 0x9D, 0x00, 0xD8, 0xBD, 0x00, 0x11, 0x9D,
      0x00, 0xD9, 0xBD, 0x00, 0x12, 0x9D, 0x00, 0xDA,
      0xBD, 0x00, 0x13, 0x9D, 0x00, 0xDB, 0xE8, 0xD0,
      0xE5,0xF0,0xFE
    };
    unsigned char fbasic[] =
    { 0x01, 0x08, 0x13, 0x08, 0x00, 0x00, 0x9e, 0x32, 0x30, 0x36, 0x39, 0x20, 0x45, 0x4e, 0x54, 0x48,
      0x55, 0x53, 0x49, 0x00, 0x00, 0x00, 0xa9, 0x00, 0x8d, 0x20, 0xd0, 0xa9, 0x00, 0x8d, 0x21, 0xd0,
      0xa9, 0x00, 0x8d, 0x22, 0xd0, 0xa9, 0x00, 0x8d, 0x23, 0xd0, 0xa9, 0x00, 0x8d, 0x16, 0xd0, 0xa9,
      0xa8, 0x8d, 0x18, 0xd0, 0xa0, 0x12, 0xa2, 0x00, 0xbd, 0x4c, 0x0c, 0x9d, 0x00, 0x20, 0xca, 0xd0,
      0xf7, 0xee, 0x39, 0x08, 0xee, 0x3c, 0x08, 0x88, 0xd0, 0xec, 0xa0, 0x04, 0xa2, 0x00, 0xbd, 0x64,
      0x08, 0x9d, 0x00, 0xd8, 0xca, 0xd0, 0xf7, 0xee, 0x4f, 0x08, 0xee, 0x52, 0x08, 0x88, 0xd0, 0xec,
      0xa9, 0x00, 0x4c, 0x61, 0x08
    };
    unsigned char pbasic[] = {
      0x01, 0x08, 0x0b, 0x08, 0xef, 0x00, 0x9e, 0x32, 0x30, 0x36, 0x31, 0x00,
      0x00, 0x00, 0xa9, 0x0b, 0x8d, 0x11, 0xd0, 0xa9, 0x14, 0x8d, 0x18, 0xd0,
      0xa9, 0x0e, 0x8d, 0x20, 0xd0, 0xa9, 0x0c, 0x8d, 0x21, 0xd0, 0xa2, 0x00,
      0xa0, 0xfa, 0xbd, 0x61, 0x08, 0x9d, 0x00, 0x04, 0xbd, 0x5b, 0x09, 0x9d,
      0xfa, 0x04, 0xbd, 0x55, 0x0a, 0x9d, 0xf4, 0x05, 0xbd, 0x4f, 0x0b, 0x9d,
      0xee, 0x06, 0xbd, 0x49, 0x0c, 0x9d, 0x00, 0xd8, 0xbd, 0x43, 0x0d, 0x9d,
      0xfa, 0xd8, 0xbd, 0x3d, 0x0e, 0x9d, 0xf4, 0xd9, 0xbd, 0x37, 0x0f, 0x9d,
      0xee, 0xda, 0xe8, 0x88, 0xd0, 0xcc, 0xa9, 0x1b, 0x8d, 0x11, 0xd0, 0x4c,
      0x5e, 0x08
    };

    FILE *vsf_file=NULL;
    FILE *map_file=NULL;
    FILE *col_file=NULL;
    FILE *prg_file=NULL;
    FILE *att_file=NULL;
    FILE *chr_file=NULL;
    FILE *scr_file=NULL;
    FILE *koa_file=NULL;
    FILE *ish_file=NULL;
    FILE *ddp_file=NULL;

    unsigned char value_d016, value_d018, value_d020, value_d021, value_dd00, byte, v1, v2, v3, v4;
    unsigned char value_d011,value_d022,value_d023;
    int i,j, bank, map, font,bitmap, file_end, pos_d018=0, pos_d800=0, pos_dd00=0;
    int hmflag=0,bmflag=0,pos_mem=0,petscii_flag=0;

    char bitmap_name[FILENAMELENGTH];
    char colmap_name[FILENAMELENGTH];
    char attrib_name[FILENAMELENGTH];
    char prg_name[FILENAMELENGTH];
    char koa_name[FILENAMELENGTH];
    char ish_name[FILENAMELENGTH];
    char dd_name[FILENAMELENGTH];
    char screen_name[FILENAMELENGTH];
    char font_name[FILENAMELENGTH];

    if (argc < 2)
    {
        LINE;
        printf (
            "VSF Screen/HiRes/Multi grabber "VERSION" by enthusi/onslaught, iAN CooG/HVSC & TSM\n"
            "Save VICE snapshot during display of any Screen.\n"
            "Original code by enthusi; Support for MultiColor Bitmap by iAN CooG;\n"
            "Doodle format & PETSCII editor-friendly saving by TSM\n"
            "Syntax: '%s ViceSnapshotFile.vsf'.\n",
            argv[0]);
        LINE;
        return -1;
    }

    sprintf (bitmap_name, "%s.bitmap", argv[1]);
    sprintf (colmap_name, "%s.colmap", argv[1]);
    sprintf (attrib_name, "%s.attrib", argv[1]);
    sprintf (koa_name, "%s.koa", argv[1]);
    sprintf (ish_name, "%s.ip64h", argv[1]);
    sprintf (dd_name, "%s.dd", argv[1]);
    sprintf (prg_name, "%s.prg", argv[1]);

    sprintf (font_name, "%s.font", argv[1]);
    sprintf (screen_name, "%s.screen", argv[1]);
    
    if (!(vsf_file = fopen (argv[1], "rb")))
    {
        printf ("Cant't open snapshot file '%s'!\n", argv[1]);
        return -1;
    }

    /* iAN: minimal sanity check */
    v1 = getc (vsf_file);
    v2 = getc (vsf_file);
    v3 = getc (vsf_file);
    v4 = getc (vsf_file);
    if (!((v1 == 'V') && (v2 == 'I') && (v3 == 'C') && (v4 == 'E')))
    {
        printf ("Not a VICE VSF file '%s'!\n", argv[1]);
        fclose(vsf_file);
        return -1;
    }

    fseek (vsf_file, 0, SEEK_END);
    file_end = ftell (vsf_file);

    for (i = 0; i < file_end; i++)
    {
        fseek (vsf_file, i, SEEK_SET);
        v1 = getc (vsf_file);
        v2 = getc (vsf_file);
        v3 = getc (vsf_file);
        v4 = getc (vsf_file);
        if( !pos_mem && (v1 == 'C') && (v2 == '6') && (v3 == '4') && (v4 == 'M'))
        {
            pos_mem = i+MEMOFFSET;
        }
        if( !pos_d018 && (v1 == 'V') && (v2 == 'I') && (v3 == 'C') && (v4 == '-'))
        {
            v1 = getc (vsf_file);
            v2 = getc (vsf_file);
            v3 = getc (vsf_file);
            v4 = getc (vsf_file);
            /* the string "VIC-II:" is found in TFR 0.8 rom */
            if( (v1 == 'I') && (v2 == 'I') && (v3 == 0) && (v4 == 0) )
            {
                for(j=0;j<10;j++)
                    getc (vsf_file);
                v1 = getc (vsf_file);
                v2 = getc (vsf_file);
                if((v1==0xe3)&&(v2==0x04))
                {
                    pos_d018 = i + VICOFFSET;
                    pos_d800 = i + ATTOFFSET;
                }
            }
        }
        if( !pos_dd00 && (v1 == 'C') && (v2 == 'I') && (v3 == 'A') && (v4 == '2'))
        {
            for(j=0;j<14;j++)
                getc (vsf_file);
            v1 = getc (vsf_file);
            if( v1 >= 0x42 )
                pos_dd00 = i + CIAOFFSET;
        }
    }
    if( !pos_mem || !pos_d018 || !pos_dd00 )
    {
        printf ("Not a VICE VSF file '%s'!\n", argv[1]);
        fclose(vsf_file);
        return -2;
    }

    fseek (vsf_file, pos_dd00, SEEK_SET);
    value_dd00 = getc (vsf_file);

    fseek (vsf_file, pos_d018-7, SEEK_SET);
    value_d011 = getc (vsf_file);
    bmflag = ((value_d011&0x20)>>5); /* 0 text, 1 gfx */

    fseek (vsf_file, pos_d018-2, SEEK_SET);
    value_d016 = getc (vsf_file);
    hmflag = (value_d016&0x10)>>4; /* 0 hires, 1 multi */

    fseek (vsf_file, pos_d018, SEEK_SET);
    value_d018 = getc (vsf_file);

    fseek (vsf_file, 7, SEEK_CUR);
    value_d020 = getc (vsf_file);
    value_d021 = getc (vsf_file);
    value_d022 = getc (vsf_file);
    value_d023 = getc (vsf_file);

    bank = 0xc000 - (value_dd00 & 3) * 0x4000;
    bitmap = ((value_d018 >> 3) & 1) * 0x2000 + bank;
    map = (value_d018 >> 4) * 0x0400 + bank;
    font = ((value_d018 >> 1) & 0x7)*0x800 + bank;
    /* prg file needed in any case */
    if (!(prg_file = fopen (prg_name, "wb")))
    {
        printf ("Cant't write PRG file '%s'!\n", prg_name);
        fclose(vsf_file);
        fclose(map_file);
        fclose(col_file);
        return -1;
    }

    /* is the snapshot in bitmap-mode after all? */
    if (bmflag)
    {
        if (!(map_file = fopen (bitmap_name, "wb")))
        {
            printf ("Cant't write bitmap file '%s'!\n", bitmap_name);
            fclose(vsf_file);
            return -1;
        }

        if(hmflag==1) /* multicolor */
        {
            if (!(koa_file = fopen (koa_name, "wb")))
            {
                printf ("Cant't write picture file '%s'!\n", koa_name);
                fclose(vsf_file);
                return -1;
            }
        }
        else /* hires */
        {
            if (!(ish_file = fopen (ish_name, "wb")))
            {
                printf ("Cant't write picture file '%s'!\n", ish_name);
                fclose(vsf_file);
                return -1;
            }
            if (!(ddp_file = fopen (dd_name, "wb")))
            {
                printf ("Cant't write picture file '%s'!\n", dd_name);
                fclose(ddp_file);
                return -1;
            }
        }

        if (!(col_file = fopen (colmap_name, "wb")))
        {
            printf ("Cant't write colmap file '%s'!\n", colmap_name);
            fclose(vsf_file);
            fclose(map_file);
            return -1;
        }

        if(hmflag) /* multicolor */
        {
            if (!(att_file = fopen (attrib_name, "wb")))
            {
                printf ("Cant't write color attributes file '%s'!\n", attrib_name);
                fclose(vsf_file);
                fclose(map_file);
                fclose(col_file);
                fclose(prg_file);
                return -1;
            };
        }

        LINE;
        printf ("Using %s Bitmap in Bank $%04x ($%02x).\n", (hmflag?"MC":"Hires"), bank, value_dd00);
        printf ("BitMap  at $%04x\twritten as '%s' without header.\n", bitmap,bitmap_name);
        printf ("ColMap  at $%04x ($%02x)\twritten as '%s' without header.\n", map,value_d018, colmap_name);
        if(hmflag)
        {
            printf ("ColAttr at $d800\twritten as '%s' without header.\n", attrib_name);
            printf ("Koala format picture\twritten as '%s'    at $6000 (BG=$%02x).\n", koa_name,value_d021);
        }
        if(bmflag & !hmflag)
        {
            printf ("HiRes format picture\twritten as '%s'  at $4000.\n", ish_name);
            printf ("HiRes format picture\twritten as '%s'  at $5C00.\n", dd_name);
        }
        printf ("Executable %s-Viewer\twritten as '%s'.\n",(hmflag?"MC":"Hires"),prg_name);
        LINE;

        /* write executable with BASIC header */
        if(hmflag)
        {
            mbasic[M_PATCH_D020]=value_d020;
            mbasic[M_PATCH_D021]=value_d021;
            fwrite (&mbasic, sizeof (mbasic), 1, prg_file);
            for (i = 0; i < 0x401-sizeof (mbasic); i++)
            {
                putc (0, prg_file);
            }
        }
        else
        {
            basic[B_PATCH_D020]=value_d020;
            fwrite (&basic, sizeof (basic), 1, prg_file);
            for (i = 0; i < 0x401-sizeof (basic); i++)
            {
                putc (0, prg_file);
            }
        }

        /* write color map data */
        fseek (vsf_file, pos_mem + map, SEEK_SET);
        for (i = 0; i < 1000; i++)
        {
            byte = getc (vsf_file);
            putc (byte, col_file);
            putc (byte, prg_file);
        }
        for (i = 0; i < 0x18; i++)
        {
            putc (0, prg_file);
        }

        /* write d800 data if needed */

        if(hmflag)
        {
            fseek (vsf_file, pos_d800, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file); /* byte &=0x0f; not needed */
                putc (byte, att_file);
                putc (byte, prg_file);
            }
            fclose(att_file);
            for (i = 0; i < (0x1000-1000); i++)
            {
                putc (0, prg_file);
            }
        }
        else
        {
            for (i = 0; i < 0x1000; i++)
            {
                putc (0, prg_file);
            }
        }

        /* write bitmap data */

        fseek (vsf_file, pos_mem + bitmap, SEEK_SET);
        for (i = 0; i < 8000; i++)
        {
            byte = getc (vsf_file);
            putc (byte, map_file);
            putc (byte, prg_file);
        }


        if(hmflag)
        {
            /* write koala picture */
            byte=0x00;  //header lsb
            putc (byte, koa_file);
            byte=0x60;  //header msb
            putc (byte, koa_file);
            fseek (vsf_file, pos_mem + bitmap, SEEK_SET);
            for (i = 0; i < 8000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, koa_file);
            }
            fseek (vsf_file, pos_mem + map, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, koa_file);
            }
            fseek (vsf_file, pos_d800, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file); /* byte &=0x0f; not needed */
                putc (byte, koa_file);
            }
            byte=value_d021;  //global BG color (C0)
            putc (byte, koa_file);
            fclose(koa_file);
        }

        if(bmflag&&!hmflag)
        {
            /* write hires picture */
            
            /* ish */
            byte=0x00;  //header lsb
            putc (byte, ish_file);
            byte=0x40;  //header msb
            putc (byte, ish_file);
            fseek (vsf_file, pos_mem + bitmap, SEEK_SET);
            for (i = 0; i < 8000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, ish_file);
            }
            fseek (vsf_file, pos_mem + map, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, ish_file);
            }
            fclose(ish_file);
            
            /* doodle (TSM) */
            byte=0x00;  //header lsb
            putc (byte, ddp_file);
            byte=0x5C;  //header msb
            putc (byte, ddp_file);
            fseek (vsf_file, pos_mem + map, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, ddp_file);
            }
            for (i = 0; i < 24; i++) putc (0x00, ddp_file); //dummy bytes
            fseek (vsf_file, pos_mem + bitmap, SEEK_SET);
            for (i = 0; i < 8000; i++)
            {
                byte = getc (vsf_file);
                putc (byte, ddp_file);
            }
            for (i = 0; i < 192; i++) putc (0x00, ddp_file);//dummy bytes
            fclose(ddp_file);
        }

        fclose(vsf_file);
        fclose(map_file);
        fclose(col_file);
        fclose(prg_file);
    }

    /* not in bitmap-mode (d011, bit5)? then the following applies: */

    if (!bmflag)
    {
        petscii_flag = (!hmflag) && is_chargen(font);
        if (!(scr_file = fopen (screen_name, "wb")))
        {
            printf ("Cant't write screen file '%s'!\n", screen_name);
            fclose(vsf_file);
            fclose(prg_file);
            return -1;
        }
        if (!(chr_file = fopen (font_name, "wb")))
        {
            printf ("Cant't write font file '%s'!\n", font_name);
            fclose(vsf_file);
            fclose(scr_file);
            fclose(prg_file);
            return -1;
        }
        if (!(att_file = fopen (attrib_name, "wb")))
        {
            printf ("Cant't write color attributes file '%s'!\n", attrib_name);
            fclose(vsf_file);
            fclose(scr_file);
            fclose(chr_file);
            fclose(prg_file);
            return -1;
        };

        /* write basic header */
        fbasic[F_PATCH_D020]=value_d020;
        fbasic[F_PATCH_D021]=value_d021;
        fbasic[F_PATCH_D022]=value_d022;
        fbasic[F_PATCH_D023]=value_d023;
        fbasic[F_PATCH_D016]=value_d016;
        pbasic[P_PATCH_D020]=value_d020;
        pbasic[P_PATCH_D021]=value_d021;
        if(is_chargen(font))
        {
            /* fix d018 to NOT use the stored font, and use chargen instead */
            fbasic[F_PATCH_D016+5]=((font&0x0800)?0xa6:0xa4);
        }

        if (petscii_flag)
        {
            fwrite (&pbasic, sizeof (pbasic), 1, prg_file);
        }
        else
        {
            fwrite (&fbasic, sizeof (fbasic), 1, prg_file);
        }

        /* write d800 data */
        fseek (vsf_file, pos_d800, SEEK_SET);
        for (i = 0; i < 1000; i++)
        {
            byte = getc (vsf_file); /* byte &=0x0f; not needed */
            putc (byte, att_file);
            if (!petscii_flag) putc (byte, prg_file);
        }

        /* write font data */
        fseek (vsf_file, pos_mem + font, SEEK_SET);
        for (i = 0; i < 0x0800; i++)
        {
            byte = getc (vsf_file);
            putc (byte, chr_file);
            if (!petscii_flag) putc (byte, prg_file);
        }

        /* write screen data */
        fseek (vsf_file, pos_mem + map, SEEK_SET);
        for (i = 0; i < 1000; i++)
        {
            byte = getc (vsf_file);
            putc (byte, scr_file);
            putc (byte, prg_file); 
        }
        
        /* write d800 data for PETSCII */
        if (petscii_flag)
        {
            fseek (vsf_file, pos_d800, SEEK_SET);
            for (i = 0; i < 1000; i++)
            {
                byte = getc (vsf_file) & 0x0f;
                putc (byte, prg_file);
            }
        }
        
        fclose(att_file);
        fclose(scr_file);
        fclose(chr_file);
        fclose(prg_file);
        fclose(vsf_file);

        LINE;
        printf ("Using screen ram in Bank $%04x ($%02x).\n", bank, value_dd00);
        printf ("Charset at $%04x%s\twritten as '%s' without header.\n", font,(is_chargen(font)?" (CHARGEN)":"\t"),font_name);
        printf ("Text screen at $%04x ($%02x)\twritten as '%s' without header.\n", map,value_d018, screen_name);
        printf ("ColAttr at $d800\t\twritten as '%s' without header.\n", attrib_name);
        printf ("Executable screen-Viewer\twritten as '%s'.\n",prg_name);
        LINE;

    }
    return 0;
}
